@@ -128,7 +128,7 @@ struct sdhci_s {
 
 // SDHCI timeouts
 #define SDHCI_POWER_OFF_TIME   1
-#define SDHCI_POWER_ON_TIME    1
+#define SDHCI_POWER_ON_TIME    5
 #define SDHCI_CLOCK_ON_TIME    1 // 74 clock cycles
 #define SDHCI_POWERUP_TIMEOUT  1000
 #define SDHCI_PIO_TIMEOUT      1000  // XXX - this is just made up
@@ -266,7 +266,7 @@ static int
 sdcard_readwrite(struct disk_op_s *op, int iswrite)
 {
     struct sddrive_s *drive = container_of(
-        op->drive_gf, struct sddrive_s, drive);
+        op->drive_fl, struct sddrive_s, drive);
     int cmd = iswrite ? SC_WRITE_SINGLE : SC_READ_SINGLE;
     if (op->count > 1)
         cmd = iswrite ? SC_WRITE_MULTIPLE : SC_READ_MULTIPLE;
